/**************************************************************
***************************************************************
*                                                             *
* All relevant (system) settings are done in configurations.h *
*                                                             *
***************************************************************
**************************************************************/


/****************************************************************************
 Include Section
*****************************************************************************/

#include "configurations.h"
#if defined DATA_CACHE_EN

#include "exception.h"

#include <cplb.h>
#include <def_LPBlackfin.h>
#include "ndef_LPBlackfin.h"

#include "macros.h"
#include "BfMemMap.h"

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __dcplb_init;
.GLOBAL __dcplb_handler;
//.GLOBAL _dcplb_handler.cplb_miss;
//.GLOBAL _dcplb_handler.protection_violation;
.GLOBAL __dflushall;
.GLOBAL __dflushinvall;

/*****************************************************************************
 Global Variables
******************************************************************************/

.GLOBAL _b4_dcplb_data_pointer;
.GLOBAL _b4_dcplb_address_pointer;

/*****************************************************************************
 Variables
******************************************************************************/

.SECTION cplb_data;

.ALIGN 4;
.BYTE4 _b4_dcplb_data_pointer = 0;
.BYTE4 _b4_dcplb_address_pointer = 0;

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

/* L1 Data Cache Bank Select */
/* Standard definition does not work well */
#define         CPLB_DCBS               0x10

/* number of entries in _b4_dcplbs_table is depending on the PAGE_SIZE */
#if defined LIMIT32kB
    #define     PAGE_SIZE_VAL           PAGE_SIZE_1KB
#else
    #define     PAGE_SIZE_VAL           PAGE_SIZE_4MB
#endif

#if   PAGE_SIZE_VAL == PAGE_SIZE_4MB
#define         PAGE_LENGTH             0x00400000
#elif PAGE_SIZE_VAL == PAGE_SIZE_1MB
#define         PAGE_LENGTH             0x00200000
#elif PAGE_SIZE_VAL == PAGE_SIZE_4KB
#define         PAGE_LENGTH             0x00001000
#elif PAGE_SIZE_VAL == PAGE_SIZE_1KB
#define         PAGE_LENGTH             0x00000400
#else
#error PAGE_SIZE undefined
#endif

/* missing definitions */
#define  CPLB_MEM_LEV 0x00000200
#define nCPLB_MEM_LEV 0x0


/***************************************************************************************
L1 Data Memory Control Register (DMEM_CONTROL)
RESERVED        [0]     1
ENDCPLB         [1]     Data CPLB enable
           DISDCPLB     0 - CLPBs disabled, minimal address checking only
            ENDCPLB     1 - CPLBs enabled
DMC           [3:2]     Enable/Disable Data Cache functionality
        ASRAM_BSRAM     00 - Both data banks are SRAM, also invalidates all cache lines if previously configured as cache
                        01 - Reserved
       ACACHE_BSRAM     10 - Data bank A is lower 16K byte SRAM, upper 16K byte cache Data bank B is SRAM
      ACACHE_BCACHE     11 - Both data banks are lower 16K byte SRAM, upper 16K byte cache
RESERVED    [11: 5]
DCBS            [4]     L1 Data Cache Bank Select
                        Valid only when DMC[1:0] = 11. Determines whether Address bit A[14] or A[23] is used to select the L1 data cache bank
         nCPLB_DCBS     0 - A[14] (16kB switch)
          CPLB_DCBS     1 - A[23] (8MB switch)
PORT_PREF0     [12]     DAG0 Port Preference
        nPORT_PREF0     0 - DAG0 non-cacheable fetches use port A
         PORT_PREF0     1 - DAG0 non-cacheable fetches use port B
PORT_PREF1     [13]     DAG1 Port Preference
        nPORT_PREF1     0 - DAG1 non-cacheable fetches use port A
         PORT_PREF1     1 - DAG1 non-cacheable fetches use port B
RESERVED    [31:14]
***************************************************************************************/

#define DMEM_CONTROL_RESET_VAL  0x1 /* Preserve Reset Status */

#if defined (DATA_CACHE_AB) && defined (DATA_CACHE_DCBS)
#define DMEM_CONTROL_VAL        ( ENDCPLB | ACACHE_BCACHE |  CPLB_DCBS | PORT_PREF0 | DMEM_CONTROL_RESET_VAL ) /* DMEM_CONTROL = 0x0000 101F */
#endif

#if defined (DATA_CACHE_AB) && !defined (DATA_CACHE_DCBS)
#define DMEM_CONTROL_VAL        ( ENDCPLB | ACACHE_BCACHE | nCPLB_DCBS | PORT_PREF0 | DMEM_CONTROL_RESET_VAL ) /* DMEM_CONTROL = 0x0000 100F */
#endif

#if !defined DATA_CACHE_AB
#define DMEM_CONTROL_VAL        ( ENDCPLB | ACACHE_BSRAM  | nCPLB_DCBS | PORT_PREF0 | DMEM_CONTROL_RESET_VAL ) /* DMEM_CONTROL = 0x0000 100B */
#endif


/***************************************************************************************
DCPLB Data Registers (DCPLB_DATAx)
CPLB_VALID              [0]
               CPLB_INVALID     0 - Invalid (disabled) CPLB entry
                 CPLB_VALID     1 - Valid (enabled) CPLB entry
CPLB_LOCK               [1]     Can be used by software in CPLB replacement algorithms
                CPLB_UNLOCK     0 - Unlocked, CPLB entry can be replaced
                  CPLB_LOCK     1 - Locked, CPLB entry should not be replaced
CPLB_USER_RD            [2]
              nCPLB_USER_RD     0 - User mode read access generates protection violation exception
               CPLB_USER_RD     1 - User mode read access permitted
CPLB_USER_WR            [3]
              nCPLB_USER_WR     0 - User mode write access generates protection violation exception
               CPLB_USER_WR     1 - User mode write access permitted
CPLB_SUPV_WR            [4]
              nCPLB_SUPV_WR     0 - Supervisor mode write access generates protection violation exception
               CPLB_SUPV_WR     1 - Supervisor mode write access permitted
RESERVED            [ 6: 5]
CPLB_DIRTY              [7]     Valid only if write back cacheable (CPLB_VALID = 1, CPLB_WT = 0, and CPLB_L1_CHBL = 1)
                 CPLB_CLEAN     0 - Clean
                 CPLB_DIRTY     1 - Dirty
                                A protection violation exception is generated on store accesses to this page when this bit is 0.
                                The state of this bit is modified only by writes to this register.
                                The exception service routine must set this bit.
RESERVED            [11: 8]
CPLB_MEM_LEV            [9]     Determines line fill and write buffers. For derivatives with internal L2 memory only.
                                The state of this bit has no affect if this page describes L1 memory.
              nCPLB_MEM_LEV     0 - High priority (usually best for on-chip L2 memory pages)
               CPLB_MEM_LEV     1 - Low priority (usually best for off-chip L2 memory pages)
CPLB_L1_CHBL           [12]     Clear this bit whenever L1 memory is configured as SRAM
              nCPLB_L1_CHBL     0 - Non-cacheable in L1
               CPLB_L1_CHBL     1 - Cacheable in L1
RESERVED               [13]
CPLB_WT                [14]     Operates only in cache mode
                    CPLB_WB     0 - Write back
                    CPLB_WT     1 - Write through
CPLB_L1_AOW            [15]     Valid only if write through cacheable (CPLB_VALID = 1, CPLB_WT = 1)
               nCPLB_L1_AOW     0 - Allocate cache lines on reads only
                CPLB_L1_AOW     1 - Allocate cache lines on reads and writes
PAGE_SIZE           [17:16]
              PAGE_SIZE_1KB     00 - 1K byte page size
              PAGE_SIZE_4KB     01 - 4K byte page size
              PAGE_SIZE_1MB     10 - 1M byte page size
              PAGE_SIZE_4MB     11 - 4M byte page size
RESERVED            [31:18]
***************************************************************************************/

/********************************************************************************
* There are two default CPLB descriptors for data accesses to the scratchpad    *
* data memory and to the system and core MMR space. These default descriptors   *
* define the above space as non-cacheable, so that additional CPLBs do not need *
* to be set up for these regions of memory.                                     *
* If valid CPLBs are set up for this space, the default CPLBs are ignored.      *
********************************************************************************/

/* DCPLB_DATA = 0x0003 0013 */
/* 4MB starting at 0xFF80 0000 covers til 0xFFC0 0000 */
/* 1MB starting at 0xFF80 0000 covers til 0xFF90 0000 -> not enough !!! */
#define DCPLB_DATA_L1_VAL                       ( CPLB_VALID | PAGE_SIZE_4MB | CPLB_LOCK   |  CPLB_USER_RD |  CPLB_USER_WR | CPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )

#if defined (DATA_CACHE_WT) && !defined (DATA_CACHE_AOW)
    /* DCPLB_DATA = 0x0003 501D */  
    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | PAGE_SIZE_VAL | CPLB_UNLOCK |  CPLB_USER_RD |  CPLB_USER_WR | CPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL | CPLB_WT | nCPLB_L1_AOW )
#endif

#if defined (DATA_CACHE_WT) && defined (DATA_CACHE_AOW)
    /* DCPLB_DATA = 0x0003 D01D */
    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | PAGE_SIZE_VAL | CPLB_UNLOCK |  CPLB_USER_RD |  CPLB_USER_WR | CPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL | CPLB_WT |  CPLB_L1_AOW )
#endif

#if !defined (DATA_CACHE_WT)
    /* DCPLB_DATA = 0x0003 109D */
    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | PAGE_SIZE_VAL | CPLB_UNLOCK |  CPLB_USER_RD |  CPLB_USER_WR |  CPLB_SUPV_WR | CPLB_DIRTY |  CPLB_L1_CHBL | CPLB_WB | nCPLB_L1_AOW )
    /* Test Purposes: DCPLB_DATA = 0x0003 101D */
//    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | PAGE_SIZE_VAL | CPLB_UNLOCK |  CPLB_USER_RD |  CPLB_USER_WR |  CPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL | CPLB_WB | nCPLB_L1_AOW )
#endif

#if defined (__ADSPBF54x__) || defined (__ADSPBF561__)
    /* DCPLB_DATA = 0x0002 101D */
    #define DCPLB_DATA_L2_SRAM_CACHE_VAL       ( CPLB_VALID | PAGE_SIZE_1MB | CPLB_LOCK |  CPLB_USER_RD |  CPLB_USER_WR |  CPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL )
    /* DCPLB_DATA = 0x0002 001D */
    #define DCPLB_DATA_L2_SRAM_NOCACHE_VAL     ( CPLB_VALID | PAGE_SIZE_1MB | CPLB_LOCK |  CPLB_USER_RD |  CPLB_USER_WR |  CPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )
#endif

#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__)
    /* DCPLB_DATA = 0x0002 1005 */
    #define DCPLB_DATA_BOOTROM_CACHE_VAL        ( CPLB_VALID | PAGE_SIZE_1MB | CPLB_LOCK   |  CPLB_USER_RD | nCPLB_USER_WR | nCPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL )
    #define DCPLB_DATA_BOOTROM_NOCACHE_VAL      ( CPLB_VALID | PAGE_SIZE_1MB | CPLB_LOCK   |  CPLB_USER_RD | nCPLB_USER_WR | nCPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )
#endif
#if defined __ADSPBF54x__
    /* DCPLB_DATA = 0x0001 1005 */
    #define DCPLB_DATA_BOOTROM_CACHE_VAL        ( CPLB_VALID | PAGE_SIZE_4KB | CPLB_LOCK   |  CPLB_USER_RD | nCPLB_USER_WR | nCPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL )
    #define DCPLB_DATA_BOOTROM_NOCACHE_VAL      ( CPLB_VALID | PAGE_SIZE_4KB | CPLB_LOCK   |  CPLB_USER_RD | nCPLB_USER_WR | nCPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )
#endif

/* ADSP-BF54x can handle more than 4MB of Async Mem !!! */
/* DCPLB_DATA = 0x0003 1005 */
#define DCPLB_DATA_ASYNC_CACHE_VAL              ( CPLB_VALID | PAGE_SIZE_4MB | CPLB_UNLOCK |  CPLB_USER_RD | nCPLB_USER_WR | CPLB_SUPV_WR | CPLB_CLEAN |  CPLB_L1_CHBL )
/* DCPLB_DATA = 0x0003 0005 */
#define DCPLB_DATA_ASYNC_NOCACHE_VAL            ( CPLB_VALID | PAGE_SIZE_4MB | CPLB_UNLOCK |  CPLB_USER_RD | nCPLB_USER_WR | CPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )

/* DCPLB_DATA = 0x0003 001D */
#define DCPLB_DATA_SDRAM_NOCACHE_VAL            ( CPLB_VALID | PAGE_SIZE_VAL | CPLB_UNLOCK |  CPLB_USER_RD |  CPLB_USER_WR |  CPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )


#define DCPLB_DATA_SDRAM_NULLPTR_PROT_VAL       ( CPLB_VALID | PAGE_SIZE_1KB | CPLB_LOCK   |  CPLB_USER_RD |  CPLB_USER_WR |  CPLB_SUPV_WR | CPLB_CLEAN | nCPLB_L1_CHBL )


/*****************************************************************************
 DCPLB Data
******************************************************************************/

.SECTION cplb_data;
.ALIGN 4;


_dcplbs_table:

.BYTE4 _b4_dcplbs_table[]=

    0xFF800000,DCPLB_DATA_L1_VAL,


    #if defined (__ADSPBF54x__) || defined (__ADSPBF561__)
//        L2_SRAM_START,DCPLB_DATA_L2_SRAM_CACHE_VAL,
        L2_SRAM_START,DCPLB_DATA_L2_SRAM_NOCACHE_VAL,
    #endif


#if ( SDRAM_SIZE == 128 || SDRAM_SIZE == 64 || SDRAM_SIZE == 32 || SDRAM_SIZE == 16 ) || defined ( LIMIT32kB )

    #if defined NULLPTR_PROTECTION
        SDRAM_START+PAGE_LENGTH*0,DCPLB_DATA_SDRAM_NULLPTR_PROT_VAL,
        SDRAM_START+PAGE_LENGTH*0,DCPLB_DATA_SDRAM_CACHE_VAL|CPLB_LOCK,
    #else
        SDRAM_START+PAGE_LENGTH*0,DCPLB_DATA_SDRAM_CACHE_VAL,
    #endif
    SDRAM_START+PAGE_LENGTH*1,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*2,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*3,DCPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 16MB */
    /* --> 4kB */

#endif


#if ( SDRAM_SIZE == 128 || SDRAM_SIZE == 64 || SDRAM_SIZE == 32 ) || defined ( LIMIT32kB )

    SDRAM_START+PAGE_LENGTH*4,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*5,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*6,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*7,DCPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 32MB */
    /* --> 8kB */

#endif


#if ( SDRAM_SIZE == 128 || SDRAM_SIZE == 64 ) || defined ( LIMIT32kB )

    SDRAM_START+PAGE_LENGTH*8,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*9,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*10,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*11,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*12,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*13,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*14,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*15,DCPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 64MB */
    /* --> 16kB */

#endif


#if ( SDRAM_SIZE == 128 ) || defined ( LIMIT32kB )

    SDRAM_START+PAGE_LENGTH*16,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*17,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*18,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*19,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*20,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*21,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*22,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*23,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*24,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*25,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*26,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*27,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*28,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*29,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*30,DCPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*31,DCPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 128MB */
    /* --> 32kB */

#endif


    #if defined (__ADSPBF51x__) || defined (__ADSPBF52x__) || defined (__ADSPBF54x__)
        BOOT_ROM_START,DCPLB_DATA_BOOTROM_CACHE_VAL,
//        BOOT_ROM_START,DCPLB_DATA_BOOTROM_NOCACHE_VAL,
    #endif


//    ASYNC_MEM_START,DCPLB_DATA_ASYNC_CACHE_VAL,
    ASYNC_MEM_START,DCPLB_DATA_ASYNC_NOCACHE_VAL,


    /* end of section - termination */
    -1;

_dcplbs_table.END:


/*****************************************************************************
 DCPLB Code
******************************************************************************/

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION cplb_code;

/*****************************************************************************
 Function:              __dcplb_init
 Description:           initialize data CPLBs
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        ASTAT,R7:5,P5:4,DAG2:0
 Global Variables Used: _b4_dcplb_address_pointer
                        _b4_dcplb_data_pointer
 C-Callable :           no
*****************************************************************************/

__dcplb_init:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:5,P5:4);
    PUSH_DAG_REGS(0);
    PUSH_DAG_REGS(1);
    PUSH_DAG_REGS(2);
    IMM32(P5,CORE_MMR_BASE);


    M0 = 0(z); M1 = 0(z); M2 = 0(z);
    L0 = 0(z); L1 = 0(z); L2 = 0(z);
    B0 = 0(z); B1 = 0(z); B2 = 0(z);

    /* RESET DMEM_CONTROL register */
    R7 = 0x1 (z);
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;

    IMM32(I0,DCPLB_ADDR0);
    IMM32(P4,_b4_dcplb_address_pointer); /* save address pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I0;
    [P4] = R6;

    IMM32(I1,DCPLB_DATA0);
    IMM32(P4,_b4_dcplb_data_pointer); /* save data pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I1;
    [P4] = R6;

    IMM32(I2,_b4_dcplbs_table);
    R5 = -1; /* end point comparison */
    R6 = 15; /* max counter */

    /* read entries from table */
    read_daddr:
        R7 = [I2++];
        CC = R7 == R5;
        if CC jump ddone;
        [I0++] = R7;
        csync;
    read_ddata:
        R7 = [I2++];
        [I1++] = R7;
        csync;
        R6 = R6 + R5;
        CC = R6 == R5;
        if !CC jump read_daddr;
    ddone:

    /* ENABLE DCPLBs */
    R7 = DMEM_CONTROL_VAL;
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;


    dcplb_init.done:
    POP_DAG_REGS(2);
    POP_DAG_REGS(1);
    POP_DAG_REGS(0);
    (R7:5,P5:4) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;
__dcplb_init.END:


/*****************************************************************************
 Function:              __dcplb_handler
 Description:           genereric data CPLB handler
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        ASTAT,R7:4
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__dcplb_handler:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:4);

    cli R4;

    R7 = SEQSTAT;
    R6 = EX_BITS;
    R5 = R7 & R6;

    /**********************
    * Verify a DCPLB miss *
    **********************/
    R6 = EX_SYS_DCPLBMISS;
    CC = R5 == R6;
    if CC jump dcplb_handler.cplb_miss (BP);

    /*************************************************
    * Verify a Data access CPLB protection violation *
    *************************************************/
    R6 = EX_SYS_DCPLBPROT;
    CC = R5 == R6;
    if CC jump dcplb_handler.protection_violation (BP);

    EMUEXCPT;
    jump(0);


    dcplb_handler.cplb_miss:
    call _dcplb_handler.cplb_miss;
    jump dcplb_handler.done;

    dcplb_handler.protection_violation:
    call _dcplb_handler.protection_violation;
//    jump dcplb_handler.done;


    dcplb_handler.done:
    bitclr(R4,bitpos(EVT_IVG14));
    sti R4;
    (R7:4) = [SP++];
    ASTAT = [SP++];
    unlink;
    rti;
__dcplb_handler.end:


/*****************************************************************************
 Function:              dcplb_handler.cplb_miss
 Description:           data access CPLB handler
                        Identify correct DCPLB entry in DCPLB table
                        Replace one (Round Robin Scheduling) with the correct entry
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        ASTAT,R7:1,P5:3,DAG1:0
 Global Variables Used: _b4_dcplb_address_pointer
                        _b4_dcplb_data_pointer
 C-Callable :           no
*****************************************************************************/

_dcplb_handler.cplb_miss:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:1,P5:3);
    PUSH_DAG_REGS(0);
    PUSH_DAG_REGS(1);
    IMM32(P5,CORE_MMR_BASE);


    /*************************************
    * Check, that access was done for    *
    * an available/populated memory area *
    *************************************/

    R6 = [P5 + lo(DCPLB_FAULT_ADDR)];      /* DCPLB_FAULT_ADDR */
    IMM32(R7,SDRAM_LENGTH);
    CC = R6 < R7;
    if CC jump dcplb_handler.cplb_miss.dcplb_exchange (BP);
    IMM32(R7,ASYNC_MEM_START);
    CC = R6 < R7;
    if !CC jump dcplb_handler.cplb_miss.dcplb_exchange (BP);

    dcplb_handler.cplb_miss.reserved_mem_access:
    EMUEXCPT;
    jump (0);

    dcplb_handler.cplb_miss.dcplb_exchange:
    /***********************************************
    * Load starting point in _b4_dcplbs_table for: *
    * DCPLB_ADDR -> I0                             *
    * DCPLB_DATA -> I1                             *
    ***********************************************/

    IMM32(I0,_b4_dcplbs_table); /* DCPLB_ADDRx */
    B0 = I0;
    M0 = 4;
    L0 = 4*length(_b4_dcplbs_table)-4;


    /*****************************************************
    * Search for the right entry in the _b4_dcplbs_table *
    * depending on DCPLB_FAULT_ADDR                      *
    *****************************************************/

    R6 = [P5 + lo(DCPLB_FAULT_ADDR)];      /* DCPLB_FAULT_ADDR */
    IMM32(R1,0x00030000); /* for PAGE_SIZE bits */

    dcplb_handler.find_DCPLB_entry:
    R5 = [I0++];    /* DCPLB_ADDRx */
    R4 = [I0++];    /* DCPLB_DATAx */
    /* Extract PAGE_SIZE bits */
    R3 = R4 & R1;
    R3 = R4 >> 16;  /* R3 = CPLB Page Size bits */


    /*****************
    * find PAGE_SIZE *
    *****************/

    CC = R3 == 0x3;
    if CC jump MB4 (BP);    /* PAGE_SIZE = 4MB */
    CC = R3 == 0x2;
    if CC jump MB1;         /* PAGE_SIZE = 1MB */
    CC = R3 == 0x1;
    if CC jump kB4;         /* PAGE_SIZE = 4kB */
                            /* else: PAGE_SIZE = 1kB */

    kB1:                    /* 1kB: 0x0400 */
    R2 = 0x400;
    jump dcplb_handler.page_size_End;

    kB4:                    /* 4kB: 0x1000 */
    R2 = 0x1000;
    jump dcplb_handler.page_size_End;

    MB1:                    /* 1MB: 0x100000 */
    IMM32(R2,0x00100000);
    jump dcplb_handler.page_size_End;

    MB4:                    /* 4MB: 0x400000 */
    IMM32(R2,0x00400000);

    dcplb_handler.page_size_End:
    R7 = R5 + R2;           /* Store DCPLB_ADDRx end address */


    /***********************************************************
    * Check if actual entry in _b4_dcplbs_table is valid       *
    * DCPLB_ADDRx_START <= DCPLB_FAULT_ADDR < DCPLB_ADDRx_END  *
    * R5 <= R6 < R7                                            *
    ***********************************************************/

    CC = R5 <= R6;
    if !CC jump dcplb_handler.find_DCPLB_entry (BP);
    CC = R6 < R7;
    if !CC jump dcplb_handler.find_DCPLB_entry (BP);


    /*************************************************
    * Load last index point in _b4_dcplbs_table for: *
    * _b4_dcplb_address_pointer -> I2                *
    * _b4_dcplb_data_pointer -> I3                   *
    * both need to be updated at the end             *
    * initialization in: __dcplb_init                *
    *************************************************/

    IMM32(B0,DCPLB_ADDR0); /* 0xFFE0 0200 */
    IMM32(P4,_b4_dcplb_address_pointer);
    R7 = [P4];
    I0 = R7;
    M0 = 4;                     /* 4 byte steps */
    L0 = 64;                    /* 16 DCPLB_ADDRx entrties with each 4 bytes */

    IMM32(B1,DCPLB_DATA0); /* 0xFFE0 0100 */
    IMM32(P3,_b4_dcplb_data_pointer);
    R7 = [P3];
    I1 = R7;
    M1 = M0;                    /* 4 byte steps */
    L1 = L0;                    /* 16 DCPLB_DATAx entrties with each 4 bytes */


    /****************************************
    * Check if actual DCPLB entry is locked *
    * If yes than try next one              *
    ****************************************/

    dcplb_handler.check_DCPLB_LOCK:
    R2 = [I0++];                /* DCPLB_ADDRx */
    NOP;
    R3 = [I1++];                /* DCPLB_DATAx */
    CC = bittst(R3,1);
    if CC jump dcplb_handler.check_DCPLB_LOCK;
    

    /* decrement index because the pointer is already at the next entry */
    R2 = [I0--];                /* DCPLB_ADDRx */
    R3 = [I1--];                /* DCPLB_DATAx */


    /**********************
    * Replace DCPLB entry *
    **********************/

    /* Before loading new descriptor data into any CPLBs, the corresponding group
    of sixteen CPLBs must be disabled using the Enable DCPLB (ENDCPLB) bit in the
    DMEM_CONTROL register for data descriptors */
    R7 = [P5 + lo(DMEM_CONTROL)];
    bitclr (R7,bitpos(ENDCPLB));
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;

    /* Update DCPLB entry and store address in _b4_dcplb_address_pointer and _b4_dcplb_data_pointer for Round Robin scheduling */
    [I0++] = R5;
    csync;
    R6 = I0;
    [P4] = R6;
    [I1++] = R4;
    csync;
    R6 = I1;
    [P3] = R6;

    /* ENABLE DCPLBs */
//    R7 = DMEM_CONTROL_VAL;
    bitset (R7,bitpos(ENDCPLB));
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;


    IMM32(P5,IMASK);
    R7 = [P5];
    bitclr (R7,bitpos(EVT_IVG14));
    [P5] = R7;


    dcplb_handler.cplb_miss.done:
    POP_DAG_REGS(1);
    POP_DAG_REGS(0);
    (R7:1,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;
_dcplb_handler.cplb_miss.end:


/*****************************************************************************
 Function:              _dcplb_handler.protection_violation
 Description:           dcplb protection violation handler
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        (none)
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

_dcplb_handler.protection_violation:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    EMUEXCPT;
    jump (0);

    dcplb_handler.protection_violation.done:
    (R7:7,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;
_dcplb_handler.protection_violation.end:


/*****************************************************************************
 Function:              __dflushall
 Description:           Flush data cache
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        P5:4,LOOPREG
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__dflushall:
    link 0;
    [--SP] = (P5:4);
    PUSH_LOOP_REGS(0);

    IMM32(P4,SDRAM_LENGTH/32);
    IMM32(P5,SDRAM_START);
    lsetup(DOFLUSHA,DOFLUSHA) LC0 = P4;
    DOFLUSHA: FLUSH[P5++];

    dflushall.done:
    POP_LOOP_REGS(0);
    (P5:4) = [SP++];
    unlink;
    rts;
__dflushall.end:


/*****************************************************************************
 Function:              __dflushinvall
 Description:           Flush and invalidate the complete data cache
                        By clearing the DMC[1:0] bits in the DMEM_CONTROL register,
                        all Valid bits in the data cache are set to the invalid state.
                        A second write to the DMEM_CONTROL register to set the
                        DMC[1:0] bits to their previous state then configures the
                        data memory back to its previous cache/SRAM configuration.
                        An SSYNC instruction should be run before invalidating the
                        cache and a CSYNC instruction should be inserted after each
                        of these operations. This method is useful if the data buffer
                        to be invalidated is greater than the size of the cache.
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        P5,R7:6
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__dflushinvall:
    link 0;
    [--SP] = (R7:6,P5:5);

    ssync; /* ensures that previous stores have been pushed through all the queues */
    IMM32(P5,DMEM_CONTROL);
    R7 = [P5];
    R6 = R7;
    bitclr(R7,2); /* DMC0 */
    bitclr(R7,3); /* DMC1 */
    [P5] = R7;
    ssync;
    [P5] = R6;
    ssync; /*   If software requires synchronization with system hardware,
                place an SSYNC instruction after the FLUSH instruction to
                ensure that the flush operation has completed */

    dflushinv.done:
    (R7:6,P5:5) = [SP++];
    unlink;
    rts;
__dflushinvall.end:


#endif /* defined DATA_CACHE_EN */


/*****************************************************************************
 EOF
******************************************************************************/
